/*
 * Copyright (C) 2003 Sistina Software.
 * Copyright (C) 2004-2005 Red Hat, Inc. All rights reserved.
 *
 * Module Author: Heinz Mauelshagen
 *
 * This file is released under the GPL.
 *
 * Round-robin path selector.
 */

#include "dm.h"
#include "dm-path-selector.h"

#include <linux/slab.h>

#define DM_MSG_PREFIX "multipath queue_depth"

/*-----------------------------------------------------------------
 * Path-handling code, paths are held in lists
 *---------------------------------------------------------------*/
struct path_info {
	struct list_head list;
	struct path *path;
	unsigned repeat_count;
	unsigned long weight;
	long long ios_done;
};

static void free_paths(struct list_head *paths)
{
	struct path_info *pi, *next;

	list_for_each_entry_safe(pi, next, paths, list) {
		list_del(&pi->list);
		kfree(pi);
	}
}

/*-----------------------------------------------------------------
 * Queue-depth selector
 *---------------------------------------------------------------*/

#define QD_MIN_IO		1000

struct selector {
	struct list_head valid_paths;
	struct list_head invalid_paths;

};

static struct selector *alloc_selector(void)
{
	struct selector *s = kmalloc(sizeof(*s), GFP_KERNEL);

	if (s) {
		INIT_LIST_HEAD(&s->valid_paths);
		INIT_LIST_HEAD(&s->invalid_paths);
	}

	return s;
}

static int qd_create(struct path_selector *ps, unsigned argc, char **argv)
{
	struct selector *s;

	s = alloc_selector();
	if (!s)
		return -ENOMEM;

	ps->context = s;
	return 0;
}

static void qd_destroy(struct path_selector *ps)
{
	struct selector *s = (struct selector *) ps->context;

	free_paths(&s->valid_paths);
	free_paths(&s->invalid_paths);
	kfree(s);
	ps->context = NULL;
}

static int qd_status(struct path_selector *ps, struct path *path,
		     status_type_t type, char *result, unsigned int maxlen)
{
	struct path_info *pi;
	int sz = 0;

	if (!path)
		DMEMIT("0 ");
	else {
		switch(type) {
		case STATUSTYPE_INFO:
			break;
		case STATUSTYPE_TABLE:
			pi = path->pscontext;
			DMEMIT("%u ", pi->repeat_count);
			break;
		}
	}

	return sz;
}

/*
 * Called during initialisation to register each path with an
 * optional repeat_count.
 */
static int qd_add_path(struct path_selector *ps, struct path *path,
		       int argc, char **argv, char **error)
{
	struct selector *s = (struct selector *) ps->context;
	struct path_info *pi;
	unsigned repeat_count = QD_MIN_IO;

	if (argc > 1) {
		*error = "queue_depth ps: incorrect number of arguments";
		return -EINVAL;
	}

	/* First path argument is number of I/Os before switching path */
	if ((argc == 1) && (sscanf(argv[0], "%u", &repeat_count) != 1)) {
		*error = "queue_depth ps: invalid repeat count";
		return -EINVAL;
	}

	/* allocate the path */
	pi = kmalloc(sizeof(*pi), GFP_KERNEL);
	if (!pi) {
		*error = "queue_depth ps: Error allocating path context";
		return -ENOMEM;
	}

	pi->path = path;
	pi->repeat_count = repeat_count;
	pi->weight  = 0 ;
	pi->ios_done = 0;
	path->pscontext = pi;
	
	list_add_tail(&pi->list, &s->valid_paths);

	return 0;
}

static void qd_fail_path(struct path_selector *ps, struct path *p)
{
	struct selector *s = (struct selector *) ps->context;
	struct path_info *pi = p->pscontext;

	list_move(&pi->list, &s->invalid_paths);
}

static int qd_reinstate_path(struct path_selector *ps, struct path *p)
{
	struct selector *s = (struct selector *) ps->context;
	struct path_info *pi = p->pscontext;

	list_move(&pi->list, &s->valid_paths);

	return 0;
}

static struct path *qd_select_path(struct path_selector *ps,
				   unsigned *repeat_count)
{
	struct selector *s = (struct selector *) ps->context;
	unsigned long min_wait = 0;
	unsigned long max_weight = 0;
	unsigned long real_max_ios = 0;
	struct path_info *pi = NULL;

	if (!list_empty(&s->valid_paths)) {
		long long max_ios;
		struct path_info *l_pi = NULL;
				
		l_pi = list_entry(s->valid_paths.next, struct path_info, list);

		/* spin lock should be added here to guard aginst start/end io  
			and maybe add path / fail path
		 */
		
		min_wait = l_pi->weight;
		max_weight = 0;
		real_max_ios = 0;
		max_ios = 0;
		list_for_each_entry(l_pi,&s->valid_paths,list) {
			if (l_pi->weight <= min_wait ) {
				if ((l_pi->ios_done > max_ios || l_pi->weight <min_wait)) {
					max_ios = l_pi->ios_done;
				}
				min_wait = l_pi->weight;
			}
			if (l_pi->weight > max_weight) {
				max_weight = l_pi->weight;
			}
			if (l_pi->ios_done > real_max_ios) {
				real_max_ios = l_pi->ios_done;
			}
		}
			
		{
			list_for_each_entry(l_pi,&s->valid_paths,list) {
				if ((l_pi->weight ==  min_wait ) && (l_pi->ios_done == max_ios) ) {
					pi = l_pi;
					list_move_tail(&pi->list, &s->valid_paths);
					break;
				}
			}
		}
		
		*repeat_count = 1;
	}

	return pi ? pi->path : NULL;
}


static int qd_start_io(struct path_selector *ps, struct path *p,unsigned int size)
{
        struct path_info *pi = p->pscontext;

	pi->weight += size;

	return 0;
}

static int qd_end_io(struct path_selector *ps, struct path *p,unsigned int size)
{
	struct selector *s = (struct selector *) ps->context;
        struct path_info *pi = p->pscontext;

	pi->weight -= size;
	pi->ios_done += 1;
	if (pi->ios_done >= 10000) {	
		if (!list_empty(&s->valid_paths)) {	
			struct path_info *l_pi = NULL;
	       		list_for_each_entry(l_pi,&s->valid_paths,list) {
				l_pi->ios_done = 0;
			}
		}
	} 
	return 0;
}

static struct path_selector_type qd_ps = {
	.name = "queue-depth",
	.module = THIS_MODULE,
	.table_args = 1,
	.info_args = 0,
	.create = qd_create,
	.destroy = qd_destroy,
	.status = qd_status,
	.add_path = qd_add_path,
	.fail_path = qd_fail_path,
	.reinstate_path = qd_reinstate_path,
	.select_path = qd_select_path,
	.start_io = qd_start_io,
	.end_io = qd_end_io,
};

static int __init dm_qd_init(void)
{
	int r = dm_register_path_selector(&qd_ps);

	if (r < 0)
		DMERR("register failed %d", r);

	DMINFO("version 1.0.0 loaded");

	return r;
}

static void __exit dm_qd_exit(void)
{
	int r = dm_unregister_path_selector(&qd_ps);

	if (r < 0)
		DMERR("queue_depth ps: unregister failed %d", r);
}

module_init(dm_qd_init);
module_exit(dm_qd_exit);

MODULE_DESCRIPTION(DM_NAME " queue_depth multipath path selector");
MODULE_AUTHOR("Exanet");
MODULE_LICENSE("GPL");
